/*
 * IPVS         An implementation of the IP virtual server support for the
 *              LINUX operating system.  IPVS is now implemented as a module
 *              over the NetFilter framework. IPVS can be used to build a
 *              high-performance and highly available server based on a
 *              cluster of servers.
 *
 * Version:     $Id: ip_vs_ctl.c,v 1.30.2.3 2003/07/29 14:37:12 wensong Exp $
 *
 * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
 *              Peter Kese <peter.kese@ijs.si>
 *              Julian Anastasov <ja@ssi.bg>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 * Changes:
 *
 */

/*
  *	Port IPVS to FreeBSD
  *	Authors:		Li Wang <dragonfly@linux-vs.org>
  *
  */

#include "ip_vs.h"

/* semaphore for IPVS sockopts. And, [gs]etsockopt may sleep. */
static struct mtx __ip_vs_mutex;

/* lock for service table */
struct sx __ip_vs_svc_lock;

/* lock for table with the real services */
static struct sx __ip_vs_rs_lock;

/* number of virtual services */
static int ip_vs_num_services = 0;

/* sysctl variables */
int sysctl_ip_vs_cache_bypass = 0;
int sysctl_ip_vs_expire_nodest_conn = 0;
int sysctl_ip_vs_sync_threshold = 3;

#ifdef CONFIG_IP_VS_DEBUG
static int sysctl_ip_vs_debug_level = 0;

int ip_vs_get_debug_level(void)
{
	return sysctl_ip_vs_debug_level;
}
#endif

/*
 *  Hash table: for virtual service lookups
 */
#define IP_VS_SVC_TAB_BITS 8
#define IP_VS_SVC_TAB_SIZE (1 << IP_VS_SVC_TAB_BITS)
#define IP_VS_SVC_TAB_MASK (IP_VS_SVC_TAB_SIZE - 1)

/* the service table hashed by <protocol, addr, port> */
static LIST_HEAD(ip_vs_svc_list, ip_vs_service) ip_vs_svc_table[IP_VS_SVC_TAB_SIZE];
/* the service table hashed by fwmark */
static LIST_HEAD(ip_vs_svc_fwm_list, ip_vs_service) ip_vs_svc_fwm_table[IP_VS_SVC_TAB_SIZE];

/*
 *  Hash table: for real service lookups
 */
#define IP_VS_RTAB_BITS 4
#define IP_VS_RTAB_SIZE (1 << IP_VS_RTAB_BITS)
#define IP_VS_RTAB_MASK (IP_VS_RTAB_SIZE - 1)

static LIST_HEAD(ip_vs_rs_list, ip_vs_dest) ip_vs_rtable[IP_VS_RTAB_SIZE];

/*
 * Trash for destinations
 */
static LIST_HEAD(ip_vs_dest_trash_list, ip_vs_dest) ip_vs_dest_trash;

/*
 * FTP & NULL virtual service counters
 */
static volatile u_int32_t ip_vs_ftpsvc_counter = 0;
static volatile u_int32_t ip_vs_nullsvc_counter = 0;


/*
 *  Returns hash value for virtual service
 */
static __inline__ unsigned
ip_vs_svc_hashkey(unsigned proto, u_int32_t addr, u_int16_t port)
{
	register unsigned porth = ntohs(port);

	return (proto^ntohl(addr)^(porth>>IP_VS_SVC_TAB_BITS)^porth)
		& IP_VS_SVC_TAB_MASK;
}

/*
 *  Returns hash value of fwmark for virtual service lookup
 */
static __inline__ unsigned ip_vs_svc_fwm_hashkey(u_int32_t fwmark)
{
	return fwmark & IP_VS_SVC_TAB_MASK;
}

/*
 *  Hashes ip_vs_service in the ip_vs_svc_table by <proto,addr,port>
 *  or in the ip_vs_svc_fwm_table by fwmark.
 *  Should be called with locked tables.
 *  Returns bool success.
 */
static int ip_vs_svc_hash(struct ip_vs_service *svc)
{
	unsigned hash;

	if (svc->flags & IP_VS_SVC_F_HASHED) {
		IP_VS_ERR("ip_vs_svc_hash(): request for already hashed, "
			  "called from %p\n", __builtin_return_address(0));
		return 0;
	}

	if (svc->fwmark == 0) {
		/*
		 *  Hash it by <protocol,addr,port> in ip_vs_svc_table
		 */
		hash = ip_vs_svc_hashkey(svc->protocol, svc->addr, svc->port);
		LIST_INSERT_HEAD(&ip_vs_svc_table[hash], svc, s_list);
	} else {
		/*
		 *  Hash it by fwmark in ip_vs_svc_fwm_table
		 */
		hash = ip_vs_svc_fwm_hashkey(svc->fwmark);
		LIST_INSERT_HEAD(&ip_vs_svc_fwm_table[hash], svc, f_list);
	}

	svc->flags |= IP_VS_SVC_F_HASHED;
	/* increase its refcnt because it is referenced by the svc table */
	atomic_add_int(&svc->refcnt, 1);
	return 1;
}


/*
 *  Unhashes ip_vs_service from ip_vs_svc_table/ip_vs_svc_fwm_table.
 *  Should be called with locked tables.
 *  Returns bool success.
 */
static int ip_vs_svc_unhash(struct ip_vs_service *svc)
{
	if (!(svc->flags & IP_VS_SVC_F_HASHED)) {
		IP_VS_ERR("ip_vs_svc_unhash(): request for unhash flagged, "
			  "called from %p\n", __builtin_return_address(0));
		return 0;
	}

	if (svc->fwmark == 0) {
		/*
		 * Remove it from the ip_vs_svc_table table.
		 */
		LIST_REMOVE(svc, s_list);
	} else {
		/*
		 * Remove it from the ip_vs_svc_fwm_table table.
		 */
		LIST_REMOVE(svc, f_list);
	}

	svc->flags &= ~IP_VS_SVC_F_HASHED;
	atomic_subtract_int(&svc->refcnt, 1);
	return 1;
}


/*
 *  Get service by {proto,addr,port} in the service table.
 */
static __inline__ struct ip_vs_service *
__ip_vs_service_get(u_int16_t protocol, u_int32_t vaddr, u_int16_t vport)
{
	unsigned hash;
	struct ip_vs_service *svc;
	struct ip_vs_svc_list *l;

	/*
	 *	Check for "full" addressed entries
	 */
	hash = ip_vs_svc_hashkey(protocol, vaddr, vport);

	l = &ip_vs_svc_table[hash];
	LIST_FOREACH(svc, l, s_list) {
		if ((svc->addr == vaddr)
		    && (svc->port == vport)
		    && (svc->protocol == protocol)) {
			/* HIT */
			atomic_add_int(&svc->usecnt, 1);
			return svc;
		}
	}

	return NULL;
}


/*
 *  Get service by {fwmark} in the service table.
 */
static __inline__ struct ip_vs_service *__ip_vs_svc_fwm_get(u_int32_t fwmark)
{
	unsigned hash;
	struct ip_vs_service *svc;
	struct ip_vs_svc_fwm_list *l;

	/*
	 *	Check for "full" addressed entries
	 */
	hash = ip_vs_svc_fwm_hashkey(fwmark);

	l = &ip_vs_svc_fwm_table[hash];
	LIST_FOREACH(svc, l, f_list) {
		if (svc->fwmark == fwmark) {
			/* HIT */
			atomic_add_int(&svc->usecnt, 1);
			return svc;
		}
	}

	return NULL;
}

struct ip_vs_service *
ip_vs_service_get(u_int32_t fwmark, u_int16_t protocol, u_int32_t vaddr, u_int16_t vport)
{
	struct ip_vs_service *svc;

	sx_slock(&__ip_vs_svc_lock);

	/*
	 *	Check the table hashed by fwmark first
	 */
	if (fwmark && (svc = __ip_vs_svc_fwm_get(fwmark)))
		goto out;

	/*
	 *	Check the table hashed by <protocol,addr,port>
	 *	for "full" addressed entries
	 */
	svc = __ip_vs_service_get(protocol, vaddr, vport);
	if (svc == NULL
	    && protocol == IPPROTO_TCP
	    && atomic_read(&ip_vs_ftpsvc_counter)
	    && (vport == FTPDATA || ntohs(vport) >= PROT_SOCK)) {
		/*
		 * Check if ftp service entry exists, the packet
		 * might belong to FTP data connections.
		 */
		svc = __ip_vs_service_get(protocol, vaddr, FTPPORT);
	}

	if (svc == NULL
	    && atomic_read(&ip_vs_nullsvc_counter)) {
		/*
		 * Check if the catch-all port (port zero) exists
		 */
		svc = __ip_vs_service_get(protocol, vaddr, 0);
	}

  out:
	sx_sunlock(&__ip_vs_svc_lock);

	IP_VS_DBG(6, "lookup service: fwm %u %s %u.%u.%u.%u:%u %s\n",
		  fwmark, ip_vs_proto_name(protocol),
		  NIPQUAD(vaddr), ntohs(vport),
		  svc?"hit":"not hit");

	return svc;
}


static inline void
__ip_vs_bind_svc(struct ip_vs_dest *dest, struct ip_vs_service *svc)
{
	atomic_add_int(&svc->refcnt, 1);
	dest->svc = svc;
}

static inline void
__ip_vs_unbind_svc(struct ip_vs_dest *dest)
{
	struct ip_vs_service *svc = dest->svc;

	dest->svc = NULL;
	if (atomic_dec_and_test(&svc->refcnt))
		free(svc, M_IPVSBUF);
}

/*
 *  Returns hash value for real service
 */
static __inline__ unsigned ip_vs_rs_hashkey(u_int32_t addr, u_int16_t port)
{
	register unsigned porth = ntohs(port);

	return (ntohl(addr)^(porth>>IP_VS_RTAB_BITS)^porth)
		& IP_VS_RTAB_MASK;
}

/*
 *  Hashes ip_vs_dest in ip_vs_rtable by proto,addr,port.
 *  should be called with locked tables.
 *  returns bool success.
 */
static int ip_vs_rs_hash(struct ip_vs_dest *dest)
{
	unsigned hash;

	if (ENTRY_ON_LIST(dest, d_list)) {
		return 0;
	}

	/*
	 *	Hash by proto,addr,port,
	 *	which are the parameters of the real service.
	 */
	hash = ip_vs_rs_hashkey(dest->addr, dest->port);
	LIST_INSERT_HEAD(&ip_vs_rtable[hash], dest, d_list);

	return 1;
}

/*
 *  UNhashes ip_vs_dest from ip_vs_rtable.
 *  should be called with locked tables.
 *  returns bool success.
 */
static inline int ip_vs_rs_unhash(struct ip_vs_dest *dest)
{
	/*
	 * Remove it from the ip_vs_rtable table.
	 */
	if (ENTRY_ON_LIST(dest, d_list)) {
		LIST_REMOVE(dest, d_list);
		INIT_LIST_ENTRY(dest, d_list);
	}

	return 1;
}

/*
 *  Lookup real service by {proto,addr,port} in the real service table.
 */
struct ip_vs_dest *
ip_vs_lookup_real_service(u_int16_t protocol, u_int32_t daddr, u_int16_t dport)
{
	unsigned hash;
	struct ip_vs_dest *dest;
	struct ip_vs_rs_list *l;

	/*
	 *	Check for "full" addressed entries
	 *	Return the first found entry
	 */
	hash = ip_vs_rs_hashkey(daddr, dport);

	l = &ip_vs_rtable[hash];

	sx_slock(&__ip_vs_rs_lock);
	LIST_FOREACH(dest, l, d_list) {
		if ((dest->addr == daddr)
		    && (dest->port == dport)
		    && ((dest->protocol == protocol) ||
			dest->vfwmark)) {
			/* HIT */
			sx_sunlock(&__ip_vs_rs_lock);
			return dest;
		}
	}
	sx_sunlock(&__ip_vs_rs_lock);

	return NULL;
}

/*
 *  Lookup destination by {addr,port} in the given service
 */
static struct ip_vs_dest *
ip_vs_lookup_dest(struct ip_vs_service *svc, u_int32_t daddr, u_int16_t dport)
{
	struct ip_vs_dest *dest;
	struct ip_vs_dest_list *l;

	/*
	 * Find the destination for the given service
	 */
	l = &svc->destinations;
	LIST_FOREACH(dest, l, n_list) {
		if ((dest->addr == daddr) && (dest->port == dport)) {
			/* HIT */
			return dest;
		}
	}

	return NULL;
}


/*
 *  Lookup dest by {svc,addr,port} in the destination trash.
 *  The destination trash is used to hold the destinations that are removed
 *  from the service table but are still referenced by some conn entries.
 *  The reason to add the destination trash is when the dest is temporary
 *  down (either by administrator or by monitor program), the dest can be
 *  picked back from the trash, the remaining connections to the dest can
 *  continue, and the counting information of the dest is also useful for
 *  scheduling.
 */
static struct ip_vs_dest *
ip_vs_trash_get_dest(struct ip_vs_service *svc, u_int32_t daddr, u_int16_t dport)
{
	struct ip_vs_dest *dest, *tmp;
	struct ip_vs_dest_trash_list *l;

	/*
	 * Find the destination in trash
	 */
	l = &ip_vs_dest_trash;

	LIST_FOREACH_SAFE(dest, l, n_list, tmp) {
		IP_VS_DBG(3, "Destination %u/%u.%u.%u.%u:%u still in trash, "
			  "refcnt=%d\n",
			  dest->vfwmark,
			  NIPQUAD(dest->addr), ntohs(dest->port),
			  atomic_read(&dest->refcnt));
		if (dest->addr == daddr &&
		    dest->port == dport &&
		    dest->vfwmark == svc->fwmark &&
		    dest->protocol == svc->protocol &&
		    (svc->fwmark ||
		     (dest->vaddr == svc->addr &&
		      dest->vport == svc->port))) {
			/* HIT */
			return dest;
		}

		/*
		 * Try to purge the destination from trash if not referenced
		 */
		if (atomic_read(&dest->refcnt) == 1) {
			IP_VS_DBG(3, "Removing destination %u/%u.%u.%u.%u:%u "
				  "from trash\n",
				  dest->vfwmark,
				  NIPQUAD(dest->addr), ntohs(dest->port));
			LIST_REMOVE(dest, n_list);
			__ip_vs_dst_release(dest);
			__ip_vs_unbind_svc(dest);
			free(dest, M_IPVSBUF);
		}
	}
	
	return NULL;
}


/*
 *  Clean up all the destinations in the trash
 *  Called by the ip_vs_control_cleanup()
 *
 *  When the ip_vs_control_clearup is activated by ipvs module exit,
 *  the service tables must have been flushed and all the connections
 *  are expired, and the refcnt of each destination in the trash must
 *  be 1, so we simply release them here.
 */
static void ip_vs_trash_cleanup(void)
{
	struct ip_vs_dest *dest, *tmp;
	struct ip_vs_dest_trash_list *l;

	l = &ip_vs_dest_trash;

	LIST_FOREACH_SAFE(dest, &ip_vs_dest_trash, n_list, tmp) {
		LIST_REMOVE(dest, n_list);
		__ip_vs_dst_release(dest);
		__ip_vs_unbind_svc(dest);
		free(dest, M_IPVSBUF);
	}
	
}


/*
 *  Update a destination in the given service
 */
static void __ip_vs_update_dest(struct ip_vs_service *svc,
				struct ip_vs_dest *dest,
				struct ip_vs_rule_user *ur)
{
	int conn_flags;

	/*
	 *    Set the weight and the flags
	 */
	atomic_set(&dest->weight, ur->weight);

	conn_flags = ur->conn_flags | IP_VS_CONN_F_INACTIVE;

	/*
	 *    Check if local node and update the flags
	 */
	/*
	if (inet_addr_type(ur->daddr) == RTN_LOCAL) {
		conn_flags = (conn_flags & ~IP_VS_CONN_F_FWD_MASK)
			| IP_VS_CONN_F_LOCALNODE;
	}*/

	/*
	 *    Set the IP_VS_CONN_F_NOOUTPUT flag if not masquerading
	 */
	if ((conn_flags & IP_VS_CONN_F_FWD_MASK) != 0) {
		conn_flags |= IP_VS_CONN_F_NOOUTPUT;
	} else {
		/*
		 *    Put the real service in ip_vs_rtable if not present.
		 *    For now only for NAT!
		 */
		sx_xlock(&__ip_vs_rs_lock);
		ip_vs_rs_hash(dest);
		sx_xunlock(&__ip_vs_rs_lock);
	}
	atomic_set(&dest->conn_flags, conn_flags);

	/* bind the service */
	if (!dest->svc) {
		__ip_vs_bind_svc(dest, svc);
	} else {
		if (dest->svc != svc) {
			__ip_vs_unbind_svc(dest);
			__ip_vs_bind_svc(dest, svc);
		}
	}

	/* set the dest status flags */
	dest->flags |= IP_VS_DEST_F_AVAILABLE;
}


/*
 *  Create a destination for the given service
 */
static int
ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur,
	       struct ip_vs_dest **destp)
{
	struct ip_vs_dest *dest;
	unsigned atype;
	struct sockaddr_in *dst;

	EnterFunction(2);

	atype = inet_addr_type(ur->daddr);
	if (atype != RTN_UNICAST)
		return EINVAL;

	*destp = dest = (struct ip_vs_dest*)
		malloc(sizeof(struct ip_vs_dest), M_IPVSBUF, M_NOWAIT|M_ZERO);
	if (dest == NULL) {
		IP_VS_ERR("ip_vs_new_dest: kmalloc failed.\n");
		return ENOMEM;
	}
	memset(dest, 0, sizeof(struct ip_vs_dest));

	dest->protocol = svc->protocol;
	dest->vaddr = svc->addr;
	dest->vport = svc->port;
	dest->vfwmark = svc->fwmark;
	dest->addr = ur->daddr;
	dest->port = ur->dport;

	atomic_set(&dest->activeconns, 0);
	atomic_set(&dest->inactconns, 0);
	atomic_set(&dest->refcnt, 0);

	dst = (struct sockaddr_in *)&dest->dst_cache.ro_dst;
	dst->sin_family = AF_INET;
	dst->sin_len = sizeof(*dst);
	dst->sin_addr.s_addr = dest->addr;

	mtx_init(&dest->dst_lock,"destination lock",NULL,MTX_DEF);
	mtx_init(&dest->stats.lock,"destination statistics lock",NULL,MTX_DEF);
	__ip_vs_update_dest(svc, dest, ur);
	
	LeaveFunction(2);
	return 0;
}


/*
 *  Add a destination into an existing service
 */
static int ip_vs_add_dest(struct ip_vs_service *svc,
			  struct ip_vs_rule_user *ur)
{
	struct ip_vs_dest *dest;
	u_int32_t daddr = ur->daddr;
	u_int16_t dport = ur->dport;
	int ret;

	EnterFunction(2);

	if (ur->weight < 0) {
		IP_VS_ERR("ip_vs_add_dest(): server weight less than zero\n");
		return ERANGE;
	}
	/*
	 * Check if the dest already exists in the list
	 */
	dest = ip_vs_lookup_dest(svc, daddr, dport);
	if (dest != NULL) {
		IP_VS_DBG(1, "ip_vs_add_dest(): dest already exists\n");
		return EEXIST;
	}
	/*
	 * Check if the dest already exists in the trash and
	 * is from the same service
	 */
	dest = ip_vs_trash_get_dest(svc, daddr, dport);
	if (dest != NULL) {
		IP_VS_DBG(3, "Get destination %u.%u.%u.%u:%u from trash, "
			  "refcnt=%d, service %u/%u.%u.%u.%u:%u\n",
			  NIPQUAD(daddr), ntohs(dport),
			  atomic_read(&dest->refcnt),
			  dest->vfwmark,
			  NIPQUAD(dest->vaddr),
			  ntohs(dest->vport));
		__ip_vs_update_dest(svc, dest, ur);

		/*
		 * Get the destination from the trash
		 */
		LIST_REMOVE(dest, n_list);

		sx_xlock(&__ip_vs_svc_lock);

		/*
		 * Wait until all other svc users go away.
		 */
		while (atomic_read(&svc->usecnt) > 1) {};

		LIST_INSERT_HEAD(&svc->destinations, dest, n_list);
		svc->num_dests++;

		/* call the update_service function of its scheduler */
		svc->scheduler->update_service(svc);

		sx_xunlock(&__ip_vs_svc_lock);
		return 0;
	}
	/*
	 * Allocate and initialize the dest structure
	 */
	ret = ip_vs_new_dest(svc, ur, &dest);
	if (ret) {
		return ret;
	}
	/*
	 * Add the dest entry into the list
	 */
	atomic_add_int(&dest->refcnt, 1);

	sx_xlock(&__ip_vs_svc_lock);

	/*
	 * Wait until all other svc users go away.
	 */
	while (atomic_read(&svc->usecnt) > 1) {};
	LIST_INSERT_HEAD(&svc->destinations, dest, n_list);
	svc->num_dests++;
	/* call the update_service function of its scheduler */
	svc->scheduler->update_service(svc);
	sx_xunlock(&__ip_vs_svc_lock);

	LeaveFunction(2);
	return 0;
}


/*
 *  Edit a destination in the given service
 */
static int ip_vs_edit_dest(struct ip_vs_service *svc,
			   struct ip_vs_rule_user *ur)
{
	struct ip_vs_dest *dest;
	u_int32_t daddr = ur->daddr;
	u_int16_t dport = ur->dport;

	EnterFunction(2);

	if (ur->weight < 0) {
		IP_VS_ERR("ip_vs_edit_dest(): server weight less than zero\n");
		return ERANGE;
	}

	/*
	 *  Lookup the destination list
	 */
	dest = ip_vs_lookup_dest(svc, daddr, dport);
	if (dest == NULL) {
		IP_VS_DBG(1, "ip_vs_edit_dest(): dest doesn't exist\n");
		return ENOENT;
	}

	__ip_vs_update_dest(svc, dest, ur);

	/* call the update_service, because server weight may be changed */
	svc->scheduler->update_service(svc);

	LeaveFunction(2);

	return 0;
}


/*
 *  Delete a destination (must be already unlinked from the service)
 */
static void __ip_vs_del_dest(struct ip_vs_dest *dest)
{

	/*
	 *  Remove it from the d-linked list with the real services.
	 */
	sx_xlock(&__ip_vs_rs_lock);
	ip_vs_rs_unhash(dest);
	sx_xunlock(&__ip_vs_rs_lock);

	/*
	 *  Decrease the refcnt of the dest, and free the dest
	 *  if nobody refers to it (refcnt=0). Otherwise, throw
	 *  the destination into the trash.
	 */
	if (atomic_dec_and_test(&dest->refcnt)) {
		__ip_vs_dst_release(dest);
		/* simply decrease svc->refcnt here, let the caller check
		   and release the service if nobody refers to it.
		   Only user context can release destination and service,
		   and only one user context can update virtual service at a
		   time, so the operation here is OK */
		atomic_subtract_int(&dest->svc->refcnt, 1);
		free(dest, M_IPVSBUF);
	} else {
		IP_VS_DBG(3, "Moving dest %u.%u.%u.%u:%u into trash, refcnt=%d\n",
			  NIPQUAD(dest->addr), ntohs(dest->port),
			  atomic_read(&dest->refcnt));

		LIST_INSERT_HEAD(&ip_vs_dest_trash, dest, n_list);
		atomic_add_int(&dest->refcnt, 1);
	}
}


/*
 *  Unlink a destination from the given service
 */
static void __ip_vs_unlink_dest(struct ip_vs_service *svc,
				struct ip_vs_dest *dest,
				int svcupd)
{
	dest->flags &= ~IP_VS_DEST_F_AVAILABLE;

	/*
	 *  Remove it from the d-linked destination list.
	 */
	LIST_REMOVE(dest, n_list);
	svc->num_dests--;
	if (svcupd) {
		/*
		 *  Call the update_service function of its scheduler
		 */
		svc->scheduler->update_service(svc);
	}
}


/*
 *  Delete a destination server in the given service
 */
static int ip_vs_del_dest(struct ip_vs_service *svc,struct ip_vs_rule_user *ur)
{
	struct ip_vs_dest *dest;
	u_int32_t daddr = ur->daddr;
	u_int16_t dport = ur->dport;

	EnterFunction(2);

	dest = ip_vs_lookup_dest(svc, daddr, dport);
	if (dest == NULL) {
		IP_VS_DBG(1, "ip_vs_del_dest(): destination not found!\n");
		return ENOENT;
	}

	sx_xlock(&__ip_vs_svc_lock);

	/*
	 *	Wait until all other svc users go away.
	 */
	while (atomic_read(&svc->usecnt) > 1) {};

	/*
	 *	Unlink dest from the service
	 */
	__ip_vs_unlink_dest(svc, dest, 1);

	sx_xunlock(&__ip_vs_svc_lock);

	/*
	 *	Delete the destination
	 */
	__ip_vs_del_dest(dest);

	LeaveFunction(2);

	return 0;
}


/*
 *  Add a service into the service hash table
 */
static int
ip_vs_add_service(struct ip_vs_rule_user *ur, struct ip_vs_service **svc_p)
{
	int ret = 0;
	struct ip_vs_scheduler *sched;
	struct ip_vs_service *svc = NULL;

	/*
	 * Lookup the scheduler, by 'ur->sched_name'
	 */
	sched = ip_vs_scheduler_get(ur->sched_name);
	if (sched == NULL) {
		IP_VS_INFO("Scheduler module ip_vs_%s.o not found\n",
			   ur->sched_name);
		ret = ENOENT;
		goto out_mod_dec;
	}

	svc = (struct ip_vs_service*)
		malloc(sizeof(struct ip_vs_service), M_IPVSBUF, M_NOWAIT|M_ZERO);
	if (svc == NULL) {
		IP_VS_DBG(1, "ip_vs_add_service: kmalloc failed.\n");
		ret = ENOMEM;
		goto out_err;
	}
	memset(svc, 0, sizeof(struct ip_vs_service));

	svc->protocol = ur->protocol;
	svc->addr = ur->vaddr;
	svc->port = ur->vport;
	svc->fwmark = ur->vfwmark;
	svc->flags = ur->vs_flags;
	svc->timeout = ur->timeout * HZ;
	svc->netmask = ur->netmask;

	LIST_INIT(&svc->destinations);
	sx_init(&svc->sched_lock,"svc sched lock");
	mtx_init(&svc->stats.lock,"svc stats lock",NULL,MTX_DEF);

	/*
	 *    Bind the scheduler
	 */
	ret = ip_vs_bind_scheduler(svc, sched);
	if (ret) {
		sx_destroy(&svc->sched_lock);
		mtx_destroy(&svc->stats.lock);
		goto out_err;
	}

	/*
	 *    Update the virtual service counters
	 */
	if (svc->port == FTPPORT)
		atomic_add_int(&ip_vs_ftpsvc_counter, 1);
	else if (svc->port == 0)
		atomic_add_int(&ip_vs_nullsvc_counter, 1);

	/*
	 *    I'm the first user of the service
	 */
	atomic_set(&svc->usecnt, 1);
	atomic_set(&svc->refcnt, 0);

	ip_vs_num_services++;

	/*
	 *    Hash the service into the service table
	 */
	sx_xlock(&__ip_vs_svc_lock);
	ip_vs_svc_hash(svc);
	sx_xunlock(&__ip_vs_svc_lock);

	*svc_p = svc;
	return 0;

  out_err:
	if (svc)
		free(svc, M_IPVSBUF);
	ip_vs_scheduler_put(sched);
  out_mod_dec:
	return ret;
}


/*
 *	Edit a service and bind it with a new scheduler
 */
static int ip_vs_edit_service(struct ip_vs_service *svc,
			      struct ip_vs_rule_user *ur)
{
	struct ip_vs_scheduler *sched, *old_sched;
	int ret = 0;

	/*
	 * Lookup the scheduler, by 'ur->sched_name'
	 */
	sched = ip_vs_scheduler_get(ur->sched_name);
	if (sched == NULL) {
		IP_VS_INFO("Scheduler module ip_vs_%s.o not found\n",
			   ur->sched_name);
		return ENOENT;
	}

	sx_xlock(&__ip_vs_svc_lock);

	/*
	 * Wait until all other svc users go away.
	 */
	while (atomic_read(&svc->usecnt) > 1) {};

	/*
	 * Set the flags and timeout value
	 */
	svc->flags = ur->vs_flags | IP_VS_SVC_F_HASHED;
	svc->timeout = ur->timeout * HZ;
	svc->netmask = ur->netmask;

	old_sched = svc->scheduler;
	if (sched != old_sched) {
		/*
		 * Unbind the old scheduler
		 */
		if ((ret = ip_vs_unbind_scheduler(svc))) {
			old_sched = sched;
			goto out;
		}

		/*
		 * Bind the new scheduler
		 */
		if ((ret = ip_vs_bind_scheduler(svc, sched))) {
			/*
			 * If ip_vs_bind_scheduler fails, restore the old
			 * scheduler.
			 * The main reason of failure is out of memory.
			 *
			 * The question is if the old scheduler can be
			 * restored all the time. TODO: if it cannot be
			 * restored some time, we must delete the service,
			 * otherwise the system may crash.
			 */
			ip_vs_bind_scheduler(svc, old_sched);
			old_sched = sched;
		}
	}

  out:
	sx_xunlock(&__ip_vs_svc_lock);

	if (old_sched)
		ip_vs_scheduler_put(old_sched);

	return ret;
}


/*
 *  Delete a service from the service list
 *  The service must be unlinked, unlocked and not referenced!
 */
static void __ip_vs_del_service(struct ip_vs_service *svc)
{
	struct ip_vs_dest_list *l;
	struct ip_vs_dest *dest, *tmp;
	struct ip_vs_scheduler *old_sched;

	ip_vs_num_services--;

	/*
	 *    Unbind scheduler
	 */
	old_sched = svc->scheduler;
	ip_vs_unbind_scheduler(svc);

	/*
	 *    Unlink the whole destination list
	 */
	l = &svc->destinations;
	LIST_FOREACH_SAFE(dest, l, n_list, tmp) {
		__ip_vs_unlink_dest(svc, dest, 0);
		__ip_vs_del_dest(dest);
	}

	/*
	 *    Update the virtual service counters
	 */
	if (svc->port == FTPPORT)
		atomic_subtract_int(&ip_vs_ftpsvc_counter, 1);
	else if (svc->port == 0)
		atomic_subtract_int(&ip_vs_nullsvc_counter, 1);

	/*
	 *    Free the service if nobody refers to it
	 */
	if (atomic_read(&svc->refcnt) == 0) {
		sx_destroy(&svc->sched_lock);
		mtx_destroy(&svc->stats.lock);
		free(svc, M_IPVSBUF);
	}
}

/*
 *  Delete a service from the service list
 */
static int ip_vs_del_service(struct ip_vs_service *svc)
{
	if (svc == NULL)
		return EEXIST;

	/*
	 * Unhash it from the service table
	 */
	sx_xlock(&__ip_vs_svc_lock);

	ip_vs_svc_unhash(svc);

	/*
	 * Wait until all the svc users go away.
	 */
	while (atomic_read(&svc->usecnt) > 1) {};

	__ip_vs_del_service(svc);

	sx_xunlock(&__ip_vs_svc_lock);

	return 0;
}


/*
 *  Flush all the virtual services
 */
static int ip_vs_flush(void)
{
	int idx;
	struct ip_vs_service *svc, *tmp;
	struct ip_vs_svc_list *l;
	struct ip_vs_svc_fwm_list *s;

	/*
	 * Flush the service table hashed by <protocol,addr,port>
	 */
	for(idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		l = &ip_vs_svc_table[idx];
		LIST_FOREACH_SAFE(svc, l, s_list, tmp) {
			sx_xlock(&__ip_vs_svc_lock);
			ip_vs_svc_unhash(svc);
			/*
			 * Wait until all the svc users go away.
			 */
			while (atomic_read(&svc->usecnt) > 0) {};
			__ip_vs_del_service(svc);
			sx_xunlock(&__ip_vs_svc_lock);
		}
	}

	/*
	 * Flush the service table hashed by fwmark
	 */
	for(idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		s = &ip_vs_svc_fwm_table[idx];
		LIST_FOREACH_SAFE(svc, s, f_list, tmp) {
			sx_xlock(&__ip_vs_svc_lock);
			ip_vs_svc_unhash(svc);
			/*
			 * Wait until all the svc users go away.
			 */
			while (atomic_read(&svc->usecnt) > 0) {};
			__ip_vs_del_service(svc);
			sx_xunlock(&__ip_vs_svc_lock);
		}
	}

	return 0;
}


/*
 *  Zero counters in a service or all services
 */
static inline void
__ip_vs_zero_stats(struct ip_vs_stats *stats)
{
	mtx_lock(&stats->lock);
	memset(stats, 0, (char *)&stats->lock - (char *)stats);
	mtx_unlock(&stats->lock);
}

static int ip_vs_zero_service(struct ip_vs_service *svc)
{
	struct ip_vs_dest *dest;

	sx_xlock(&__ip_vs_svc_lock);

	LIST_FOREACH(dest, &svc->destinations, n_list) {
		__ip_vs_zero_stats(&dest->stats);
	}
	__ip_vs_zero_stats(&svc->stats);
	sx_xunlock(&__ip_vs_svc_lock);
	return 0;
}

static int ip_vs_zero_all(void)
{
	int idx;
	struct ip_vs_service *svc;

	for(idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		LIST_FOREACH(svc, &ip_vs_svc_table[idx], s_list) {
			ip_vs_zero_service(svc);
		}
	}

	for(idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		LIST_FOREACH(svc, &ip_vs_svc_fwm_table[idx], f_list) {
			ip_vs_zero_service(svc);
		}
	}

	__ip_vs_zero_stats(&ip_vs_stats);
	return 0;
}


/*
 *	Write the contents of the VS rule table to a PROCfs file.
 *	(It is kept just for backward compatibility)
 */
static inline char *ip_vs_fwd_name(unsigned flags)
{
	char *fwd;

	switch (flags & IP_VS_CONN_F_FWD_MASK) {
	case IP_VS_CONN_F_LOCALNODE:
		fwd = "Local";
		break;
	case IP_VS_CONN_F_TUNNEL:
		fwd = "Tunnel";
		break;
	case IP_VS_CONN_F_DROUTE:
		fwd = "Route";
		break;
	default:
		fwd = "Masq";
	}
	return fwd;
}

#define 	net_msg_cost 	5*hz
#define	net_msg_burst	10*5*hz

/* 
 * This enforces a rate limit: not more than one kernel message
 * every 5secs to make a denial-of-service attack impossible.
 *
 * All warning printf()s should be guarded by this function. 
 */ 
int net_ratelimit(void)
{
	static struct mtx ratelimit_lock;
	static unsigned long toks;
	static unsigned long last_msg; 
	static int missed;
	unsigned long now = ticks;

	if (!mtx_initialized(&ratelimit_lock)) {
		mtx_init(&ratelimit_lock,"ratelimit lock",NULL,MTX_DEF);
		toks=10*5*hz;
	}
	mtx_lock(&ratelimit_lock);
	toks += now - last_msg;
	last_msg = now;
	if (toks > net_msg_burst)
		toks = net_msg_burst;
	if (toks >= net_msg_cost) {
		int lost = missed;
		missed = 0;
		toks -= net_msg_cost;
		mtx_unlock(&ratelimit_lock);
		if (lost)
			log(LOG_INFO, "NET: %d messages suppressed.\n", lost);
		return 1;
	}
	missed++;
	mtx_unlock(&ratelimit_lock);
	return 0;
}

struct ip_vs_stats ip_vs_stats;

/*
 * Set timeout values for tcp tcpfin udp in the vs_timeout_table.
 */
static int ip_vs_set_timeouts(struct ip_vs_rule_user *u)
{
	IP_VS_DBG(2, "Setting timeout tcp:%d tcpfin:%d udp:%d\n",
		  u->tcp_timeout,
		  u->tcp_fin_timeout,
		  u->udp_timeout);

	if (u->tcp_timeout) {
		vs_timeout_table.timeout[IP_VS_S_ESTABLISHED]
			= u->tcp_timeout * hz;
	}

	if (u->tcp_fin_timeout) {
		vs_timeout_table.timeout[IP_VS_S_FIN_WAIT]
			= u->tcp_fin_timeout * hz;
	}

	if (u->udp_timeout) {
		vs_timeout_table.timeout[IP_VS_S_UDP]
			= u->udp_timeout * hz;
	}
	return 0;
}


static int
do_ip_vs_set_ctl(struct socket *sk, int cmd, void *user, unsigned int len)
{
	int ret;
	struct ip_vs_rule_user *urule;
	struct ip_vs_service *svc = NULL;

	if (suser(curthread))
		return EPERM;
	/*
	 * Check the size of mm, no overflow...
	 * len > 128000 is a sanity check.
	 */
	if (len < sizeof(struct ip_vs_rule_user)) {
		IP_VS_ERR("set_ctl: len %u < %u\n",
			  len, sizeof(struct ip_vs_rule_user));
		return EINVAL;
	} else if (len > 128000) {
		IP_VS_ERR("set_ctl: len %u > 128000\n", len);
		return EINVAL;
	} else if ((urule = malloc(len, M_IPVSBUF, M_WAITOK|M_ZERO)) == NULL) {
		IP_VS_ERR("set_ctl: no mem for len %u\n", len);
		return ENOMEM;
	} else if (copyin(user, urule, len) != 0) {
		ret = EFAULT;
		goto out_free;
	}

	mtx_lock(&__ip_vs_mutex);

	if (cmd == IP_VS_SO_SET_FLUSH) {
		/* Flush the virtual service */
		ret = ip_vs_flush();
		goto out_unlock;
	} else if (cmd == IP_VS_SO_SET_TIMEOUTS) {
		/* Set timeout values for (tcp tcpfin udp) */
		ret = ip_vs_set_timeouts(urule);
		goto out_unlock;
	} else if (cmd == IP_VS_SO_SET_ZERO) {
		/* if no service address is set, zero counters in all */
		if (!urule->vfwmark && !urule->vaddr && !urule->vport) {
			ret = ip_vs_zero_all();
			goto out_unlock;
		}
	}

	/*
	 * Check for valid protocol: TCP or UDP. Even for fwmark!=0
	 */
	if (urule->protocol!=IPPROTO_TCP && urule->protocol!=IPPROTO_UDP) {
		IP_VS_INFO("vs_ctl: invalid protocol: %d %u.%u.%u.%u:%d %s",
			   ntohs(urule->protocol), NIPQUAD(urule->vaddr),
			   ntohs(urule->vport), urule->sched_name);
		ret = EFAULT;
		goto out_unlock;
	}

	/*
	 * Lookup the exact service by <protocol, vaddr, vport> or fwmark
	 */
	if (urule->vfwmark == 0)
		svc = __ip_vs_service_get(urule->protocol,
					  urule->vaddr, urule->vport);
	else
		svc = __ip_vs_svc_fwm_get(urule->vfwmark);

	if (cmd != IP_VS_SO_SET_ADD
	    && (svc == NULL || svc->protocol != urule->protocol)) {
		ret = ESRCH;
		goto out_unlock;
	}

	switch (cmd) {
	case IP_VS_SO_SET_ADD:
		if (svc != NULL)
			ret = EEXIST;
		else
			ret = ip_vs_add_service(urule, &svc);
		break;
	case IP_VS_SO_SET_EDIT:
		ret = ip_vs_edit_service(svc, urule);
		break;
	case IP_VS_SO_SET_DEL:
		ret = ip_vs_del_service(svc);
		if (!ret)
			goto out_unlock;
		break;
	case IP_VS_SO_SET_ADDDEST:
		ret = ip_vs_add_dest(svc, urule);
		break;
	case IP_VS_SO_SET_EDITDEST:
		ret = ip_vs_edit_dest(svc, urule);
		break;
	case IP_VS_SO_SET_DELDEST:
		ret = ip_vs_del_dest(svc, urule);
		break;
	case IP_VS_SO_SET_ZERO:
		ret = ip_vs_zero_service(svc);
		break;
	default:
		ret = EINVAL;
	}

	if (svc)
		ip_vs_service_put(svc);

  out_unlock:
  	mtx_unlock(&__ip_vs_mutex);
  out_free:
  	free(urule, M_IPVSBUF);
	return ret;
}


static inline void
__ip_vs_copy_stats(struct ip_vs_stats_user *dst, struct ip_vs_stats *src)
{
	mtx_lock(&src->lock);
	bcopy(src, dst, (char*)&src->lock - (char*)src);
	mtx_unlock(&src->lock);
}

static inline int
__ip_vs_get_service_entries(const struct ip_vs_get_services *get,
			    struct ip_vs_get_services *uptr)
{
	int idx, count=0;
	struct ip_vs_service *svc;
	struct ip_vs_service_user entry;
	int ret = 0;

	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		LIST_FOREACH(svc, &ip_vs_svc_table[idx], s_list) {
			if (count >= get->num_services)
				goto out;
			entry.protocol = svc->protocol;
			entry.addr = svc->addr;
			entry.port = svc->port;
			entry.fwmark = svc->fwmark;
			strcpy(entry.sched_name, svc->scheduler->name);
			entry.flags = svc->flags;
			entry.timeout = svc->timeout / HZ;
			entry.netmask = svc->netmask;
			entry.num_dests = svc->num_dests;
			__ip_vs_copy_stats(&entry.stats, &svc->stats);
			if (copyout(&entry, &uptr->entrytable[count],
					  sizeof(entry))) {
				ret = EFAULT;
				goto out;
			}
			count++;
		}
	}

	for (idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++) {
		LIST_FOREACH(svc, &ip_vs_svc_fwm_table[idx], f_list) {
			if (count >= get->num_services)
				goto out;
			entry.protocol = svc->protocol;
			entry.addr = svc->addr;
			entry.port = svc->port;
			entry.fwmark = svc->fwmark;
			strcpy(entry.sched_name, svc->scheduler->name);
			entry.flags = svc->flags;
			entry.timeout = svc->timeout / HZ;
			entry.netmask = svc->netmask;
			entry.num_dests = svc->num_dests;
			__ip_vs_copy_stats(&entry.stats, &svc->stats);
			if (copyout(&entry, &uptr->entrytable[count],
					  sizeof(entry))) {
				ret = EFAULT;
				goto out;
			}
			count++;
		}
	}
 out:
	return ret;
}

static inline int
__ip_vs_get_dest_entries(const struct ip_vs_get_dests *get,
			 struct ip_vs_get_dests *uptr)
{
	struct ip_vs_service *svc;
	int ret = 0;

	if (get->fwmark)
		svc = __ip_vs_svc_fwm_get(get->fwmark);
	else
		svc = __ip_vs_service_get(get->protocol,
					  get->addr, get->port);
	if (svc) {
		int count = 0;
		struct ip_vs_dest *dest;
		struct ip_vs_dest_list *l;
		struct ip_vs_dest_user entry;

		l = &svc->destinations;
		LIST_FOREACH(dest, l, n_list) {
			if (count >= get->num_dests)
				break;
			entry.addr = dest->addr;
			entry.port = dest->port;
			entry.flags = atomic_read(&dest->conn_flags);
			entry.weight = atomic_read(&dest->weight);
			entry.activeconns = atomic_read(&dest->activeconns);
			entry.inactconns = atomic_read(&dest->inactconns);
			__ip_vs_copy_stats(&entry.stats, &dest->stats);
			if (copyout(&entry, &uptr->entrytable[count],
					  sizeof(entry))) {
				ret = EFAULT;
				break;
			}
			count++;
		}
		ip_vs_service_put(svc);
	} else
		ret = ESRCH;
	return ret;
}

static inline void
__ip_vs_get_timeouts(struct ip_vs_timeout_user *u)
{
	u->tcp_timeout = vs_timeout_table.timeout[IP_VS_S_ESTABLISHED] / HZ;
	u->tcp_fin_timeout = vs_timeout_table.timeout[IP_VS_S_FIN_WAIT] / HZ;
	u->udp_timeout = vs_timeout_table.timeout[IP_VS_S_UDP] / HZ;
}

static int
do_ip_vs_get_ctl(struct socket *sk, int cmd, void *user, int *len)
{
	int ret = 0;

	if (suser(curthread))
		return EPERM;

	mtx_lock(&__ip_vs_mutex);

	switch (cmd) {
	case IP_VS_SO_GET_VERSION:
	{
		char buf[64];

		sprintf(buf, "IP Virtual Server version %d.%d.%d (size=%d)",
			NVERSION(IP_VS_VERSION_CODE), IP_VS_CONN_TAB_SIZE);
		if (*len < strlen(buf)+1) {
			ret = EINVAL;
			goto out;
		}
		if (copyout(buf, user, strlen(buf)+1) != 0) {
			ret = EFAULT;
			goto out;
		}
		*len = strlen(buf)+1;
	}
	break;

	case IP_VS_SO_GET_INFO:
	{
		struct ip_vs_getinfo info;
		info.version = IP_VS_VERSION_CODE;
		info.size = IP_VS_CONN_TAB_SIZE;
		info.num_services = ip_vs_num_services;
		if (copyout(&info, user, sizeof(info)) != 0)
			ret = EFAULT;
	}
	break;

	case IP_VS_SO_GET_SERVICES:
	{
		struct ip_vs_get_services get;

		if (*len < sizeof(get)) {
			IP_VS_ERR("length: %u < %u\n", *len, sizeof(get));
			ret = EINVAL;
			goto out;
		}
		if (copyin(user, &get, sizeof(get))) {
			ret = EFAULT;
			goto out;
		}
		if (*len != (sizeof(get)+sizeof(struct ip_vs_service_user)*get.num_services)) {
			IP_VS_ERR("length: %u != %u\n", *len,
				  sizeof(get)+sizeof(struct ip_vs_service_user)*get.num_services);
			ret = EINVAL;
			goto out;
		}
		ret = __ip_vs_get_service_entries(&get, user);
	}
	break;

	case IP_VS_SO_GET_SERVICE:
	{
		struct ip_vs_service_user get;
		struct ip_vs_service *svc;

		if (*len != sizeof(get)) {
			IP_VS_ERR("length: %u != %u\n", *len, sizeof(get));
			ret = EINVAL;
			goto out;
		}
		if (copyin(user, &get, sizeof(get))) {
			ret = EFAULT;
			goto out;
		}

		if (get.fwmark)
			svc = __ip_vs_svc_fwm_get(get.fwmark);
		else
			svc = __ip_vs_service_get(get.protocol,
						  get.addr, get.port);
		if (svc) {
			strcpy(get.sched_name, svc->scheduler->name);
			get.flags = svc->flags;
			get.timeout = svc->timeout / HZ;
			get.netmask = svc->netmask;
			get.num_dests = svc->num_dests;
			__ip_vs_copy_stats(&get.stats, &svc->stats);
			if (copyout(&get, user, *len) != 0)
				ret = EFAULT;
			ip_vs_service_put(svc);
		} else
			ret = ESRCH;
	}
	break;

	case IP_VS_SO_GET_DESTS:
	{
		struct ip_vs_get_dests get;

		if (*len < sizeof(get)) {
			IP_VS_ERR("length: %u < %u\n", *len, sizeof(get));
			ret = EINVAL;
			goto out;
		}
		if (copyin(user, &get, sizeof(get))) {
			ret = EFAULT;
			goto out;
		}
		if (*len != (sizeof(get) +
			     sizeof(struct ip_vs_dest_user)*get.num_dests)) {
			IP_VS_ERR("length: %u != %u\n", *len,
				  sizeof(get)+sizeof(struct ip_vs_dest_user)*get.num_dests);
			ret = EINVAL;
			goto out;
		}
		ret = __ip_vs_get_dest_entries(&get, user);
	}
	break;

	case IP_VS_SO_GET_TIMEOUTS:
	{
		struct ip_vs_timeout_user u;

		if (*len < sizeof(u)) {
			IP_VS_ERR("length: %u < %u\n", *len, sizeof(u));
			ret = EINVAL;
			goto out;
		}
		__ip_vs_get_timeouts(&u);
		if (copyout(&u, user, sizeof(u)) != 0)
			ret = EFAULT;
	}
	break;

	default:
		ret = EINVAL;
	}

  out:
  	mtx_unlock(&__ip_vs_mutex);
	return ret;
}


static struct nf_sockopt_ops ip_vs_sockopts = {
	{ NULL, NULL }, PF_INET,
	IP_VS_BASE_CTL, IP_VS_SO_SET_MAX+1, do_ip_vs_set_ctl,
	IP_VS_BASE_CTL, IP_VS_SO_GET_MAX+1, do_ip_vs_get_ctl
};

int ip_vs_control_init(void)
{
	int ret;
	int idx;

	EnterFunction(2);

	mtx_init(&__ip_vs_mutex,"IPVS mutex",NULL,MTX_DEF);
	sx_init(&__ip_vs_svc_lock,"IPVS service lock");
	sx_init(&__ip_vs_rs_lock,"IPVS real server lock");
	LIST_INIT(&ip_vs_dest_trash);
	
	ret = nf_register_sockopt(&ip_vs_sockopts);
	if (ret) {
		IP_VS_ERR("cannot register sockopt.\n");
		return ret;
	}

	/*
	 * Initilize ip_vs_svc_table, ip_vs_svc_fwm_table, ip_vs_rtable,
	 * ip_vs_schedulers.
	 */
	for(idx = 0; idx < IP_VS_SVC_TAB_SIZE; idx++)  {
		LIST_INIT(&ip_vs_svc_table[idx]);
		LIST_INIT(&ip_vs_svc_fwm_table[idx]);
	}
	for(idx = 0; idx < IP_VS_RTAB_SIZE; idx++)  {
		LIST_INIT(&ip_vs_rtable[idx]);
	}

	memset(&ip_vs_stats, 0, sizeof(ip_vs_stats));
	mtx_init(&ip_vs_stats.lock,"ipvs stats lock",NULL,MTX_DEF);


	LeaveFunction(2);
	return 0;
}

void ip_vs_control_cleanup(void)
{
	EnterFunction(2);
	ip_vs_trash_cleanup();
	nf_unregister_sockopt(&ip_vs_sockopts);
	mtx_destroy(&ip_vs_stats.lock);
	mtx_destroy(&__ip_vs_mutex);
	sx_destroy(&__ip_vs_svc_lock);
	sx_destroy(&__ip_vs_rs_lock);
	LeaveFunction(2);
}
